part of '../machine.dart';

typedef InvokeFutureBuilderTransitionReduce<TValue, TEvent, TContext, T>
    = TContext Function(TContext context, T data);

abstract class BuilderBase<TValue, TEvent, TContext> {
  late StateBuilder<TValue, TEvent, TContext> state =
      StateBuilder<TValue, TEvent, TContext>();

  late EffectBuilder<TValue, TEvent, TContext> effect =
      EffectBuilder<TValue, TEvent, TContext>();

  Machine<TValue, TEvent, TContext> machine(
          MachineConfig<TValue, TEvent, TContext> config,
          {DebugLogger? logger}) =>
      Machine<TValue, TEvent, TContext>(config, logger: logger);

  ManagedMachine<TValue, TEvent, TContext> managed(
          MachineConfig<TValue, TEvent, TContext> config,
          {DebugLogger? logger}) =>
      ManagedMachine<TValue, TEvent, TContext>(config, logger: logger);

  MachineConfig<TValue, TEvent, TContext> config({
    required TValue initial,
    required TContext context,
    required Map<TValue, StateNodeConfig<TValue, TEvent, TContext>> states,
  }) =>
      MachineConfig<TValue, TEvent, TContext>(
          initialValue: initial, initialContext: context, states: states);

  /// Create an immediate transition.
  /// It is a transition that is taken immediately when a state is entered.
  TransitionImmediate<TValue, TEvent, TContext> immediate(
    TValue to, {
    TransitionImmediateReduce<TValue, TEvent, TContext>? reduce,
    TransitionImmediateCondition<TValue, TEvent, TContext>? condition,
  }) {
    return TransitionImmediate<TValue, TEvent, TContext>(
      to: to,
      reduce: reduce,
      condition: condition,
    );
  }
}

class StateBuilder<TValue, TEvent, TContext> {
  /// Create an atomic state using [atomic]
  StateNodeConfig<TValue, TEvent, TContext> call({
    List<TransitionImmediate<TValue, TEvent, TContext>>? immediate,
    List<TransitionEvent<TValue, TEvent, TContext>>? transition,
    List<EffectEnter<TValue, TEvent, TContext>>? effect,
    ReduceOnEnter<TValue, TEvent, TContext>? reduceOnEnter,
    ReduceOnExit<TValue, TEvent, TContext>? reduceOnExit,
    ManagedMachineConfig<TValue, TEvent, TContext>? managed,
  }) =>
      atomic(
        immediate: immediate,
        transition: transition,
        effect: effect,
        reduceOnEnter: reduceOnEnter,
        reduceOnExit: reduceOnExit,
        managed: managed,
      );

  StateNodeConfig<TValue, TEvent, TContext> atomic({
    List<TransitionImmediate<TValue, TEvent, TContext>>? immediate,
    List<TransitionEvent<TValue, TEvent, TContext>>? transition,
    List<EffectEnter<TValue, TEvent, TContext>>? effect,
    ReduceOnEnter<TValue, TEvent, TContext>? reduceOnEnter,
    ReduceOnExit<TValue, TEvent, TContext>? reduceOnExit,
    ManagedMachineConfig<TValue, TEvent, TContext>? managed,
  }) {
    return StateNodeConfig<TValue, TEvent, TContext>.atomic(
      immediate: immediate ?? const [],
      transition: transition ?? const [],
      effect: effect ?? const [],
      reduceOnEnter: reduceOnEnter,
      reduceOnExit: reduceOnExit,
      managed: managed,
    );
  }

  late final ManagedMachineBuilder<TValue, TEvent, TContext> manage =
      ManagedMachineBuilder<TValue, TEvent, TContext>();

  /// Immediately invoke a [Future] and transition on result/error.
  StateNodeConfig<TValue, TEvent, TContext>
      future<TResult, Error extends Object>({
    required Future<TResult> Function(void Function(void Function()) cleanUp)
        future,
    required TValue success,
    required TValue error,
    void Function(dynamic error)? onUnknownError,
    InvokeFutureBuilderTransitionReduce<TValue, TEvent, TContext, TResult>?
        onSuccessReduce,
    InvokeFutureBuilderTransitionReduce<TValue, TEvent, TContext, Error>?
        onErrorReduce,
    List<TransitionImmediate<TValue, TEvent, TContext>>? immediate,
  }) {
    return StateNodeConfig<TValue, TEvent, TContext>.atomic(
      immediate: immediate ?? const [],
      transition: [
        TransitionEvent<TValue, TEvent, TContext>(
            to: success,
            isCandidate: (eventType) => eventType.maybeMap(
                  onSuccess: (obj) => obj.data is TResult,
                  orElse: () => false,
                ),
            reduce: null != onSuccessReduce
                ? (ctx, e) {
                    final _event =
                        e as EventTypeOnSuccess<TValue, TEvent, TContext>;
                    return onSuccessReduce(ctx, _event.data as TResult);
                  }
                : null),
        TransitionEvent<TValue, TEvent, TContext>(
          to: error,
          isCandidate: (eventType) => eventType.maybeMap(
            onError: (obj) => obj.data is Error,
            orElse: () => false,
          ),
          reduce: null != onErrorReduce
              ? (ctx, e) {
                  final _event =
                      e as EventTypeOnError<TValue, TEvent, TContext>;
                  return onErrorReduce(ctx, _event.data as Error);
                }
              : null,
        ),
      ],
      effect: [
        (effect) {
          final List<void Function()> disposer = [];

          future(disposer.add).then(
            (value) {
              if (effect.isActive) {
                effect._machine.dispatch(
                    EventType<TValue, TEvent, TContext>.onSuccess(value));
              }
            },
          ).onError<Error>(
            (err, trace) {
              if (effect.isActive) {
                effect._machine.dispatch(
                    EventType<TValue, TEvent, TContext>.onError(err,
                        trace: trace));
              }
            },
          ).catchError(
            (dynamic err) {
              if (null != onUnknownError) {
                onUnknownError.call(err);
              } else {
                throw Exception(
                    '''Invokation of Future<${_typeOf<TResult>()}> results in Error "$err" which has no Transition defined in ${_typeOf<StateNodeConfig<TValue, TEvent, TContext>>()}
Expected an Error of type ${_typeOf<Error>()}''');
              }
            },
          );

          return (_) => disposer
            ..forEach((element) => element())
            ..clear();
        }
      ],
    );
  }
}

class EffectBuilder<TValue, TEvent, TContext> {
  EffectEnter<TValue, TEvent, TContext> call(
          EffectEnter<TValue, TEvent, TContext> cb) =>
      create(cb);

  EffectEnter<TValue, TEvent, TContext> create(
          EffectEnter<TValue, TEvent, TContext> cb) =>
      cb;

  EffectEnter<TValue, TEvent, TContext> log({
    Object Function(MachineState<TValue, TEvent, TContext> ctx)? enter,
    Object Function(MachineState<TValue, TEvent, TContext> ctx)? exit,
  }) {
    return (effect) {
      assert(() {
        print(enter?.call(effect.getState()));
        return true;
      }(), '');

      return (state) {
        assert(() {
          print(exit?.call(state));
          return true;
        }(), '');
      };
    };
  }
}

class ManagedMachineBuilder<TValue, TEvent, TContext> {
  ManagedMachineConfig<TValue, TEvent, TContext> one<AValue, AContext>({
    required ManagedMachine<AValue, Object?, AContext> machine,
    AValue Function(MachineState<TValue, TEvent, TContext> state)? value,
    AContext Function(MachineState<TValue, TEvent, TContext> state)? context,
    void Function(MachineState<AValue, Object, AContext> state,
            void Function(TEvent event) send)?
        change,
  }) {
    return ManagedMachineConfig<TValue, TEvent, TContext>(
      items: [
        ManagedMachineItemConfig<TValue, TEvent, TContext, AValue, AContext>(
            machine: machine, value: value, context: context),
      ],
      anyChange: null == change
          ? null
          : (states, send) =>
              change(states[0] as MachineState<AValue, Object, AContext>, send),
    );
  }

  ManagedMachineConfig<TValue, TEvent, TContext>
      two<AValue, AContext, BValue, BContext>({
    required ManagedMachine<AValue, Object?, AContext> machineA,
    required ManagedMachine<BValue, Object?, BContext> machineB,
    AValue Function(MachineState<TValue, TEvent, TContext> state)? valueA,
    AContext Function(MachineState<TValue, TEvent, TContext> state)? contextA,
    BValue Function(MachineState<TValue, TEvent, TContext> state)? valueB,
    BContext Function(MachineState<TValue, TEvent, TContext> state)? contextB,
    void Function(
            MachineState<AValue, Object, AContext> stateA,
            MachineState<BValue, Object, BContext> stateB,
            void Function(TEvent event) send)?
        anyChange,
  }) {
    return ManagedMachineConfig<TValue, TEvent, TContext>(
      items: [
        ManagedMachineItemConfig<TValue, TEvent, TContext, AValue, AContext>(
          machine: machineA,
          value: valueA,
          context: contextA,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, BValue, BContext>(
          machine: machineB,
          value: valueB,
          context: contextB,
        ),
      ],
      anyChange: null == anyChange
          ? null
          : (states, send) => anyChange(
              states[0] as MachineState<AValue, Object, AContext>,
              states[1] as MachineState<BValue, Object, BContext>,
              send),
    );
  }

  ManagedMachineConfig<TValue, TEvent, TContext>
      three<AValue, AContext, BValue, BContext, CValue, CContext>({
    required ManagedMachine<AValue, Object, AContext> machineA,
    required ManagedMachine<BValue, Object, BContext> machineB,
    required ManagedMachine<CValue, Object, CContext> machineC,
    AValue Function(MachineState<TValue, TEvent, TContext> state)? valueA,
    AContext Function(MachineState<TValue, TEvent, TContext> state)? contextA,
    BValue Function(MachineState<TValue, TEvent, TContext> state)? valueB,
    BContext Function(MachineState<TValue, TEvent, TContext> state)? contextB,
    CValue Function(MachineState<TValue, TEvent, TContext> state)? valueC,
    CContext Function(MachineState<TValue, TEvent, TContext> state)? contextC,
    void Function(
            MachineState<AValue, Object, AContext> stateA,
            MachineState<BValue, Object, BContext> stateB,
            MachineState<CValue, Object, CContext> stateC,
            void Function(TEvent event) send)?
        anyChange,
  }) {
    return ManagedMachineConfig<TValue, TEvent, TContext>(
      items: [
        ManagedMachineItemConfig<TValue, TEvent, TContext, AValue, AContext>(
          machine: machineA,
          value: valueA,
          context: contextA,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, BValue, BContext>(
          machine: machineB,
          value: valueB,
          context: contextB,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, CValue, CContext>(
          machine: machineC,
          value: valueC,
          context: contextC,
        ),
      ],
      anyChange: null == anyChange
          ? null
          : (states, send) => anyChange(
              states[0] as MachineState<AValue, Object, AContext>,
              states[1] as MachineState<BValue, Object, BContext>,
              states[2] as MachineState<CValue, Object, CContext>,
              send),
    );
  }

  ManagedMachineConfig<TValue, TEvent, TContext> four<AValue, AContext, BValue,
      BContext, CValue, CContext, DValue, DContext>({
    required ManagedMachine<AValue, Object, AContext> machineA,
    required ManagedMachine<BValue, Object, BContext> machineB,
    required ManagedMachine<CValue, Object, CContext> machineC,
    required ManagedMachine<DValue, Object, DContext> machineD,
    AValue Function(MachineState<TValue, TEvent, TContext> state)? valueA,
    AContext Function(MachineState<TValue, TEvent, TContext> state)? contextA,
    BValue Function(MachineState<TValue, TEvent, TContext> state)? valueB,
    BContext Function(MachineState<TValue, TEvent, TContext> state)? contextB,
    CValue Function(MachineState<TValue, TEvent, TContext> state)? valueC,
    CContext Function(MachineState<TValue, TEvent, TContext> state)? contextC,
    DValue Function(MachineState<TValue, TEvent, TContext> state)? valueD,
    DContext Function(MachineState<TValue, TEvent, TContext> state)? contextD,
    void Function(
            MachineState<AValue, Object, AContext> stateA,
            MachineState<BValue, Object, BContext> stateB,
            MachineState<CValue, Object, CContext> stateC,
            MachineState<DValue, Object, DContext> stateD,
            void Function(TEvent event) send)?
        anyChange,
  }) {
    return ManagedMachineConfig<TValue, TEvent, TContext>(
      items: [
        ManagedMachineItemConfig<TValue, TEvent, TContext, AValue, AContext>(
          machine: machineA,
          value: valueA,
          context: contextA,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, BValue, BContext>(
          machine: machineB,
          value: valueB,
          context: contextB,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, CValue, CContext>(
          machine: machineC,
          value: valueC,
          context: contextC,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, DValue, DContext>(
          machine: machineD,
          value: valueD,
          context: contextD,
        ),
      ],
      anyChange: null == anyChange
          ? null
          : (states, send) => anyChange(
              states[0] as MachineState<AValue, Object, AContext>,
              states[1] as MachineState<BValue, Object, BContext>,
              states[2] as MachineState<CValue, Object, CContext>,
              states[3] as MachineState<DValue, Object, DContext>,
              send),
    );
  }

  ManagedMachineConfig<TValue, TEvent, TContext> five<AValue, AContext, BValue,
      BContext, CValue, CContext, DValue, DContext, EValue, EContext>({
    required ManagedMachine<AValue, Object, AContext> machineA,
    required ManagedMachine<BValue, Object, BContext> machineB,
    required ManagedMachine<CValue, Object, CContext> machineC,
    required ManagedMachine<DValue, Object, DContext> machineD,
    required ManagedMachine<EValue, Object, EContext> machineE,
    AValue Function(MachineState<TValue, TEvent, TContext> state)? valueA,
    AContext Function(MachineState<TValue, TEvent, TContext> state)? contextA,
    BValue Function(MachineState<TValue, TEvent, TContext> state)? valueB,
    BContext Function(MachineState<TValue, TEvent, TContext> state)? contextB,
    CValue Function(MachineState<TValue, TEvent, TContext> state)? valueC,
    CContext Function(MachineState<TValue, TEvent, TContext> state)? contextC,
    DValue Function(MachineState<TValue, TEvent, TContext> state)? valueD,
    DContext Function(MachineState<TValue, TEvent, TContext> state)? contextD,
    EValue Function(MachineState<TValue, TEvent, TContext> state)? valueE,
    EContext Function(MachineState<TValue, TEvent, TContext> state)? contextE,
    void Function(
            MachineState<AValue, Object, AContext> stateA,
            MachineState<BValue, Object, BContext> stateB,
            MachineState<CValue, Object, CContext> stateC,
            MachineState<DValue, Object, DContext> stateD,
            MachineState<EValue, Object, EContext> stateE,
            void Function(TEvent event) send)?
        anyChange,
  }) {
    return ManagedMachineConfig<TValue, TEvent, TContext>(
      items: [
        ManagedMachineItemConfig<TValue, TEvent, TContext, AValue, AContext>(
          machine: machineA,
          value: valueA,
          context: contextA,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, BValue, BContext>(
          machine: machineB,
          value: valueB,
          context: contextB,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, CValue, CContext>(
          machine: machineC,
          value: valueC,
          context: contextC,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, DValue, DContext>(
          machine: machineD,
          value: valueD,
          context: contextD,
        ),
        ManagedMachineItemConfig<TValue, TEvent, TContext, EValue, EContext>(
          machine: machineE,
          value: valueE,
          context: contextE,
        ),
      ],
      anyChange: null == anyChange
          ? null
          : (states, send) => anyChange(
              states[0] as MachineState<AValue, Object, AContext>,
              states[1] as MachineState<BValue, Object, BContext>,
              states[2] as MachineState<CValue, Object, CContext>,
              states[3] as MachineState<DValue, Object, DContext>,
              states[4] as MachineState<EValue, Object, EContext>,
              send),
    );
  }
}
