import 'dart:async';

import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:collection/collection.dart';
import 'package:flutter/foundation.dart';

part 'state_node.dart';
part 'scheduler.dart';
part 'effect.dart';
part 'debugLog.dart';

part 'builder/builder_base.dart';
part 'builder/machine_builder.dart';

part 'machine.freezed.dart';

Type _typeOf<T>() => T;
Scheduler _globalScheduler = Scheduler()..initialize();

@freezed
class EventType<TValue, TEvent, TContext>
    with _$EventType<TValue, TEvent, TContext> {
  factory EventType.custom(TEvent event) =
      EventTypeCustom<TValue, TEvent, TContext>;

  factory EventType.onSuccess(dynamic data) =
      EventTypeOnSuccess<TValue, TEvent, TContext>;

  factory EventType.onError(dynamic data, {StackTrace? trace}) =
      EventTypeOnError<TValue, TEvent, TContext>;
}

@freezed
class MachineStatus<TValue, TEvent, TContext>
    with _$MachineStatus<TValue, TEvent, TContext> {
  factory MachineStatus.notStarted() =
      MachineStatusNotStarted<TValue, TEvent, TContext>;
  factory MachineStatus.stopped(
          {required MachineState<TValue, TEvent, TContext> lastState}) =
      MachineStatusStopped<TValue, TEvent, TContext>;
  factory MachineStatus.running(
          {required MachineState<TValue, TEvent, TContext> state}) =
      MachineStatusRunning<TValue, TEvent, TContext>;
}

@freezed
class MachineState<TValue, TEvent, TContext>
    with _$MachineState<TValue, TEvent, TContext> {
  factory MachineState({
    required TValue value,
    required TContext context,
  }) = _MachineState<TValue, TEvent, TContext>;

  static MachineState<TValue, TEvent, TContext> from<TValue, TEvent, TContext>({
    required TValue value,
    required TContext context,
  }) =>
      MachineState<TValue, TEvent, TContext>(
        value: value,
        context: context,
      );
}

class MachineConfig<TValue, TEvent, TContext> {
  MachineConfig({
    required this.initialValue,
    required this.initialContext,
    required this.states,
  });
  final TValue initialValue;
  final TContext initialContext;
  final Map<TValue, StateNodeConfig<TValue, TEvent, TContext>> states;

  MachineState<TValue, TEvent, TContext> get initial =>
      MachineState<TValue, TEvent, TContext>(
        value: initialValue,
        context: initialContext,
      );
}

abstract class MachineBase<TValue, TEvent, TContext> extends ChangeNotifier {
  MachineBase(
    this._config, {
    Scheduler? scheduler,
    DebugLogger? logger,
  })  : this.logger = logger,
        this._scheduler = scheduler ?? _globalScheduler,
        _status = MachineStatus<TValue, TEvent, TContext>.notStarted() {
    _states =
        _config.states.map((key, value) => MapEntry(key, StateNode(value)));
  }

  MachineStatus<TValue, TEvent, TContext> _status;
  MachineStatus<TValue, TEvent, TContext> get status => _status;

  final MachineConfig<TValue, TEvent, TContext> _config;
  final Scheduler _scheduler;
  final Scheduler _notifyScheduler = Scheduler()..initialize();
  final List<_MachineEffect<TValue, TEvent, TContext>> _effects = [];
  final DebugLogger? logger;
  late final Map<TValue, StateNode<TValue, TEvent, TContext>> _states;
  MachineBase<dynamic, dynamic, dynamic>? _parent;

  String get _machineType;

  /// DO NOT USE OUTSIDE OF TESTS
  /// THIS WILL ALWAYS BE NULL IN _PRODUCTION_
  TValue? get debugState {
    late TValue? result;
    assert(() {
      result = status.map(
          notStarted: (_) => null,
          stopped: (obj) => obj.lastState.value,
          running: (obj) => obj.state.value);
      return true;
    }(), '');
    return result;
  }

  /// DO NOT USE OUTSIDE OF TESTS
  /// THIS WILL ALWAYS BE NULL IN _PRODUCTION_
  TValue? get debugStateValue {
    late TValue? result;
    assert(() {
      result = status.map(
          notStarted: (_) => null,
          stopped: (obj) => obj.lastState.value,
          running: (obj) => obj.state.value);
      return true;
    }(), '');
    return result;
  }

  /// DO NOT USE OUTSIDE OF TESTS
  /// THIS WILL ALWAYS BE NULL IN _PRODUCTION_
  TContext? get debugStateContext {
    late TContext? result;
    assert(() {
      result = status.map(
          notStarted: (_) => null,
          stopped: (obj) => obj.lastState.context,
          running: (obj) => obj.state.context);
      return true;
    }(), '');
    return result;
  }

  @override
  void dispose() {
    assert(_parent == null,
        'Disposing $_machineType that is managed by ${_parent!._machineType} is not allowed.');

    status.maybeMap<void>(
      running: (obj) {
        _stop();
      },
      orElse: () {},
    );

    _parent = null;
    super.dispose();
  }

  StateNode<TValue, TEvent, TContext> _getNode(TValue value) {
    if (!_states.containsKey(value)) {
      throw Exception('There is no state of $value in $_machineType');
    }

    return _states[value] as StateNode<TValue, TEvent, TContext>;
  }

  void _notify([Scheduler? scheduler]) => status.map(
        notStarted: (_) => throw Exception(
            'A not started machine should never notifyListeners'),
        stopped: (stopped) {
          final s = scheduler ?? _notifyScheduler;
          s.schedule(() {
            notifyListeners();

            final node = _getNode(stopped.lastState.value);
            node.config.maybeMap(
              orElse: () {},
              atomic: (atomic) {
                if (null == atomic.managed) return;
                atomic.managed!.items
                    .forEach((item) => item.machine.notifyListeners());
                atomic.managed!.items
                    .forEach((item) => item.machine._notify(s));
              },
            );
          });
        },
        running: (running) {
          final s = scheduler ?? _notifyScheduler;
          s.schedule(() {
            notifyListeners();

            final node = _getNode(running.state.value);
            node.config.maybeMap(
              orElse: () {},
              atomic: (atomic) {
                if (null == atomic.managed) return;
                atomic.managed!.items
                    .forEach((item) => item.machine.notifyListeners());
                atomic.managed!.items
                    .forEach((items) => items.machine._notify(s));
              },
            );
          });
        },
      );

  MachineState<TValue, TEvent, TContext> _takeImmediateTransition(
      MachineState<TValue, TEvent, TContext> state) {
    final node = _getNode(state.value);

    final immediate = node.getImmediateCandidate(state.context);
    if (null != immediate) {
      return _takeImmediateTransition(
        state.copyWith.call(
            value: immediate.to,
            context: immediate.reduce?.call(state.context) ?? state.context),
      );
    }

    return state;
  }

  bool _canTransition(EventType<TValue, TEvent, TContext> eventType) =>
      status.map<bool>(
          notStarted: (_) => false,
          stopped: (_) => false,
          running: (obj) =>
              null !=
              _getNode(obj.state.value)
                  .getEventCandidate(eventType, obj.state.context));

  void _reduceOnLeavingState() => status.maybeMap(
        orElse: () => throw Exception(
            'Tryint to reduce while exiting invalid machine status: $status'),
        running: (running) {
          final node = _getNode(running.state.value);
          node.config.map(atomic: (atomic) {
            atomic.managed?.items
                .forEach((item) => item.machine._reduceOnLeavingState());
          });

          _status = running.copyWith.call(
            state: running.state.copyWith.call(
              context: node.config.reduceOnExit?.call(running.state.context) ??
                  running.state.context,
            ),
          );
        },
      );

  void _reduceOnEnterState() => status.maybeMap(
      orElse: () => throw Exception(
          'Tryint to reduce while entering invalid machine status: $status'),
      running: (running) {
        final node = _getNode(running.state.value);

        _status = running.copyWith.call(
          state: running.state.copyWith.call(
            context: node.config.reduceOnEnter?.call(running.state.context) ??
                running.state.context,
          ),
        );

        node.config.map(atomic: (atomic) {
          atomic.managed?.items
              .forEach((item) => item.machine._reduceOnEnterState());
        });
      });

  void _executeEffects() {
    status.maybeMap(
        orElse: () => throw Exception(
            'Trying to execute effects in invalid status $status'),
        running: (running) {
          assert(() {
            _log(Log.executeEffects(_machineType, running.state));

            return true;
          }(), '');

          _effects
            ..addAll(_getNode(running.state.value)
                .config
                .effect
                .map((cb) => _MachineEffect<TValue, TEvent, TContext>(cb)))
            ..forEach((effect) => effect.execute(this, running.state));

          _getNode(running.state.value).config.map(atomic: (atomic) {
            atomic.managed?.items
                .forEach((item) => item.machine._executeEffects());
          });
        });
  }

  void _cleanEffects() {
    status.maybeMap(
        orElse: () => throw Exception(
            'Trying to cleanup effects in invalid status $status'),
        running: (running) {
          _getNode(running.state.value).config.map(atomic: (atomic) {
            atomic.managed?.items
                .forEach((item) => item.machine._cleanEffects());
          });
          _effects
            ..forEach((effect) => effect.kill(running.state))
            ..clear();
        });
  }

  void _informManagedChangedCallback({
    bool deep = true,
  }) {
    _doInRunning((state) {
      _getNode(state.value).config.maybeMap(
          orElse: () {},
          atomic: (atomic) {
            if (null == atomic.managed) return;
            final statesList = atomic.managed!.items
                .fold<List<MachineState<dynamic, Object?, dynamic>>>([],
                    (list, item) {
              item.machine._doInRunning((state) => list.add(state));
              return list;
            });
            atomic.managed!.anyChange?.call(statesList, this.send);

            if (deep) {
              atomic.managed!.items.forEach(
                  (item) => item.machine._informManagedChangedCallback());
            }
          });
    });
  }

  void _log(Log log) {
    if (null == logger) return;
    logger!.log(log);
  }

  void _start(MachineState<TValue, TEvent, TContext> startWith) {
    status.map<void>(
      notStarted: (_) {
        _scheduler.schedule(() {
          _status = MachineStatus<TValue, TEvent, TContext>.running(
              state: _takeImmediateTransition(startWith));
          assert(() {
            _log(Log.machineStartInitially(_machineType, startWith));
            return true;
          }(), '');
          _startManaged();
          _reduceOnEnterState();
          _executeEffects();
          _informManagedChangedCallback();
          _notify();
        });
      },
      stopped: (obj) {
        _scheduler.schedule(() {
          _status = MachineStatus<TValue, TEvent, TContext>.running(
              state: _takeImmediateTransition(startWith));

          assert(() {
            _log(Log.machineStartStopped(_machineType, startWith));
            return true;
          }(), '');
          _startManaged();
          _reduceOnEnterState();
          _executeEffects();
          _informManagedChangedCallback();
          _notify();
        });
      },
      running: (_) => throw Exception(
          'You are trying to start an already started Machine $_machineType'),
    );
  }

  /// Stop the service.
  /// The serive will no longer accept any event.
  /// It will execute all exit actions in the current state.
  ///
  /// Calling this function while service is not running will result in an Exception
  void _stop() {
    status.map(
      notStarted: (_) => throw Exception(
          'You are trying to stop Machine $_machineType which has never been started'),
      stopped: (_) => throw Exception(
          'You are trying to stop an already stopped Machine  $_machineType'),
      running: (obj) {
        _reduceOnLeavingState();
        _cleanEffects();
        _stopManaged();
        _parent = null;

        _status.maybeMap(
          orElse: () => throw Exception('invalid status $status'),
          running: (running) {
            _status = MachineStatus<TValue, TEvent, TContext>.stopped(
              lastState: running.state,
            );
            assert(() {
              _log(Log.machineStop(_machineType, running.state));
              return true;
            }(), '');
          },
        );
        _notify();
      },
    );
  }

  void _stopManaged() {
    _doInRunning((state) {
      _getNode(state.value).config.map(atomic: (atomic) {
        atomic.managed?.items.forEach((item) => item.machine._stopFromParent());
      });
    });
  }

  void _startManaged() {
    _doInRunning((state) {
      _getNode(state.value).config.map(atomic: (atomic) {
        atomic.managed?.items.forEach((item) {
          item.machine._startFromParent<TValue, TEvent, TContext>(
            this,
            value: item.value?.call(state),
            context: item.context?.call(state),
          );
        });
      });
    });
  }

  void _updateRunningState(
      MachineState<TValue, TEvent, TContext> Function(
              MachineState<TValue, TEvent, TContext> state)
          cb) {
    _status.maybeMap(
      orElse: () => throw Exception(
          'Trying to update the state of $_machineType which is not running'),
      running: (running) {
        _status = running.copyWith.call(state: cb(running.state));
      },
    );
  }

  void _doInRunning(
      void Function(MachineState<TValue, TEvent, TContext> state) cb) {
    _status.maybeMap(
      orElse: () => throw Exception(
          'Expected $_machineType to be runnning which it is not.'),
      running: (running) {
        cb(running.state);
      },
    );
  }

  @internal
  void dispatch(EventType<TValue, TEvent, TContext> eventType) {
    status.map<void>(
      stopped: (_) => throw Exception(
          'You cannot send an event to stopped Machine $_machineType'),
      notStarted: (_) => throw Exception(
          'You cannot send an event to Machine $_machineType. It is not started yet'),
      running: (running) {
        assert(
          () {
            return _canTransition(eventType);
          }(),
          'There is no transition for event $eventType in state ${running.state}',
        );

        _scheduler.schedule(() {
          final candidate = _getNode(running.state.value)
              .getEventCandidate(eventType, running.state.context);

          if (null == candidate) return;

          _updateRunningState((s) => s.copyWith.call(
              context:
                  candidate.reduce?.call(s.context, eventType) ?? s.context));
          _reduceOnLeavingState();
          _cleanEffects();
          _stopManaged();

          _updateRunningState((s) =>
              _takeImmediateTransition(s.copyWith.call(value: candidate.to)));

          assert(() {
            _doInRunning((s) {
              _log(Log.transition(_machineType, running.state, s));
            });

            return true;
          }(), '');

          _startManaged();

          _reduceOnEnterState();
          _executeEffects();
          if (null != _parent) {
            _parent!._informManagedChangedCallback(deep: false);
          }
          _notify();
        });
      },
    );
  }

  /// Ask wheter an event can be send to the service.
  /// Will always return false if the service is not running.
  bool canTransition(TEvent event) => status.map<bool>(
      notStarted: (_) => false,
      stopped: (_) => false,
      running: (obj) =>
          null !=
          _getNode(obj.state.value).getEventCandidate(
              EventType<TValue, TEvent, TContext>.custom(event),
              obj.state.context));

  /// Send an event to the service which may result in a state transition
  void send(TEvent event) {
    assert(() {
      _log(Log.sendEvent(_machineType, event));
      return true;
    }(), '');
    dispatch(EventType<TValue, TEvent, TContext>.custom(event));
  }
}

class Machine<TValue, TEvent, TContext>
    extends MachineBase<TValue, TEvent, TContext> {
  Machine(MachineConfig<TValue, TEvent, TContext> config, {DebugLogger? logger})
      : super(config, logger: logger);

  @override
  String get _machineType => '${_typeOf<Machine<TValue, TEvent, TContext>>()}';

  void start({
    TValue? value,
    TContext? context,
  }) {
    _start(
      MachineState<TValue, TEvent, TContext>(
        value: value ?? _config.initialValue,
        context: context ?? _config.initialContext,
      ),
    );
  }

  void stop() => _stop();
}

class ManagedMachine<TValue, TEvent, TContext>
    extends MachineBase<TValue, TEvent, TContext> {
  ManagedMachine(MachineConfig<TValue, TEvent, TContext> config,
      {DebugLogger? logger})
      : super(config, logger: logger);

  @override
  String get _machineType =>
      '${_typeOf<ManagedMachine<TValue, TEvent, TContext>>()}';

  void _startFromParent<ParentValue, ParentEvent, ParentContext>(
    MachineBase<ParentValue, ParentEvent, ParentContext> parent, {
    Object? value,
    Object? context,
  }) {
    _parent = parent;

    assert(_status.maybeMap(running: (_) => false, orElse: () => true),
        '$_machineType is already running while trying to be started by ${parent._machineType}');
    _status.maybeMap(running: (_) => _stop(), orElse: () {});

    final initial = _config.initial;

    final state = _takeImmediateTransition(initial.copyWith.call(
      value: value is TValue ? value : initial.value,
      context: context is TContext ? context : initial.context,
    ));
    _status = MachineStatus<TValue, TEvent, TContext>.running(
      state: state,
    );

    assert(() {
      _log(Log.machineStartManaged(_machineType, parent._machineType, state));
      return true;
    }(), '');

    _getNode(state.value).config.map(atomic: (atomic) {
      atomic.managed?.items.forEach(
          (item) => item.machine._startFromParent<TValue, TEvent, TContext>(
                this,
                value: item.value?.call(state),
                context: item.context?.call(state),
              ));
    });
  }

  void _stopFromParent() {
    _doInRunning((state) {
      _getNode(state.value).config.map(atomic: (atomic) {
        atomic.managed?.items.forEach((item) => item.machine._stopFromParent());
      });
      final parentType = _parent!._machineType;
      _parent = null;
      _status = MachineStatus<TValue, TEvent, TContext>.stopped(
        lastState: state,
      );

      assert(() {
        _log(Log.machineStopManaged(_machineType, parentType, state));
        return true;
      }(), '');
    });
  }
}
