import 'dart:async';

import 'package:flutter_endlich/flutter_endlich.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:test/test.dart';

part 'builder_test.freezed.dart';

@freezed
class FooEvent with _$FooEvent {
  const factory FooEvent.Next() = _Next;
  const factory FooEvent.NextWithData(int data) = _NextWithData;
  const factory FooEvent.ToFoo() = _ToFoo;
  const factory FooEvent.ToBar() = _ToBar;
  const factory FooEvent.ToBaz() = _ToBaz;
}

enum FooState { Foo, Bar, Baz }

void main() {
  group('Builder', () {
    group('Future State', () {
      test('successfully transition', () async {
        final b = MachineBuilder<FooState, FooEvent, Null>();
        final completer = Completer<Null>();
        final service = b.machine(b.config(
          initial: FooState.Foo,
          context: null,
          states: {
            FooState.Foo: b.state.future(
              future: (_) => completer.future,
              success: FooState.Bar,
              error: FooState.Baz,
            ),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        completer.complete();
        await completer.future;

        expect(service.debugStateValue, FooState.Bar);
      });

      test('successfully transition on error', () async {
        final b = MachineBuilder<FooState, FooEvent, Null>();
        final completer = Completer<Null>();
        final service = b.machine(b.config(
          initial: FooState.Foo,
          context: null,
          states: {
            FooState.Foo: b.state.future<Null, String>(
                future: (_) => completer.future,
                success: FooState.Bar,
                error: FooState.Baz),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        completer.completeError('oh no');
        try {
          await completer.future;
        } catch (_) {}

        expect(service.debugStateValue, FooState.Baz);
      });

      test('reduce on success', () async {
        final b = MachineBuilder<FooState, FooEvent, String>();
        final completer = Completer<String>();
        final service = b.machine(b.config(
          initial: FooState.Foo,
          context: 'i am ',
          states: {
            FooState.Foo: b.state.future<String, String>(
              future: (_) => completer.future,
              success: FooState.Bar,
              error: FooState.Baz,
              onSuccessReduce: (ctx, data) => ctx + data,
            ),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        completer.complete('cool');
        await completer.future;

        expect(service.debugStateContext, 'i am cool');
      });

      test('reduce on error', () async {
        final b = MachineBuilder<FooState, FooEvent, String>();
        final completer = Completer<Null>();
        final service = b.machine(b.config(
          initial: FooState.Foo,
          context: '',
          states: {
            FooState.Foo: b.state.future<Null, String>(
                future: (_) => completer.future,
                success: FooState.Bar,
                error: FooState.Baz,
                onErrorReduce: (ctx, err) => err),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        completer.completeError('oh no');
        try {
          await completer.future;
        } catch (_) {}

        expect(service.debugStateContext, 'oh no');
      });

      test('invalid error', () async {
        final b = MachineBuilder<FooState, FooEvent, String>();
        final completer = Completer<Null>();
        dynamic err;
        b.machine(b.config(
          initial: FooState.Foo,
          context: '',
          states: {
            FooState.Foo: b.state.future<Null, String>(
              future: (_) => completer.future,
              success: FooState.Bar,
              error: FooState.Baz,
              onUnknownError: (dynamic error) {
                err = error;
              },
            ),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        completer.completeError(100);
        try {
          await completer.future;
        } catch (_, __) {}
        expect(err, 100);
      });
    });
  });
}
