import 'dart:async';

import 'package:flutter_endlich/flutter_endlich.dart';
import 'package:flutter_endlich/src/machine.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/annotations.dart';
import 'package:test/test.dart';
import 'package:mockito/mockito.dart';

import 'machine_test.mocks.dart';
part 'machine_test.freezed.dart';

class EffectEnterMock extends Mock {
  EffectLeave<FooState, FooEvent, int>? call(
      EffectValue<FooState, FooEvent, int> effect);
}

class EffectExitMock extends Mock {
  void call(MachineState<FooState, FooEvent, int> state);
}

@freezed
class FooEvent with _$FooEvent {
  const factory FooEvent.Next() = _Next;
  const factory FooEvent.NextWithData(int data) = _NextWithData;
  const factory FooEvent.ToFoo() = _ToFoo;
  const factory FooEvent.ToBar() = _ToBar;
  const factory FooEvent.ToBaz() = _ToBaz;
}

enum FooState { Foo, Bar, Baz }

@GenerateMocks([EffectEnterMock, EffectExitMock])
void main() {
  group('machine', () {
    group('status', () {
      test('has correct initial status', () {
        final b = MachineBuilder<FooState, FooEvent, void>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: null,
          states: {
            FooState.Foo: b.state(),
          },
        ));

        expect(machine.status,
            MachineStatus<FooState, FooEvent, void>.notStarted());
      });

      test('changes status', () {
        final b = MachineBuilder<FooState, FooEvent, void>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: null,
          states: {
            FooState.Foo: b.state(),
          },
        ));

        expect(machine.status,
            MachineStatus<FooState, FooEvent, void>.notStarted());

        machine.start();
        expect(machine.status, isA<MachineStatusRunning>());

        machine.stop();
        expect(machine.status, isA<MachineStatusStopped>());

        machine.start();
        expect(machine.status, isA<MachineStatusRunning>());

        machine.dispose();
        expect(machine.status, isA<MachineStatusStopped>());
      });
    });

    group('.start()', () {
      test('will start in initial state provided by config', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 5,
          states: {
            FooState.Foo: b.state(),
            FooState.Bar: b.state(),
          },
        ))
          ..start();

        expect(machine.debugStateValue, FooState.Foo);
        expect(machine.debugStateContext, 5);
      });

      test('will take immediate transition', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 5,
          states: {
            FooState.Foo: b.state(immediate: [b.immediate(FooState.Bar)]),
            FooState.Bar: b.state(),
          },
        ))
          ..start();

        expect(machine.debugStateValue, FooState.Bar);
        expect(machine.debugStateContext, 5);
      });

      test('will execute effects', () {
        final b = MachineBuilder<FooState, FooEvent, int>();
        final mock = MockEffectEnterMock();
        final mock2 = MockEffectEnterMock();
        when(mock.call(any)).thenReturn(null);
        when(mock2.call(any)).thenReturn(null);

        b.machine(b.config(
          initial: FooState.Foo,
          context: 5,
          states: {
            FooState.Foo: b.state(effect: [
              b.effect.call(mock.call),
              b.effect.call(mock2.call),
            ]),
          },
        ))
          ..start();

        verify(mock.call(any)).called(1);
        verify(mock2.call(any)).called(1);
      });

      test('with different value or context', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(),
            FooState.Bar: b.state(),
          },
        ))
          ..start(value: FooState.Bar);

        expect(machine.debugStateValue, FooState.Bar);
        expect(machine.debugStateContext, 0);

        machine
          ..stop()
          ..start(context: 1);

        expect(machine.debugStateValue, FooState.Foo);
        expect(machine.debugStateContext, 1);

        machine
          ..stop()
          ..start(value: FooState.Bar, context: 2);
        expect(machine.debugStateValue, FooState.Bar);
        expect(machine.debugStateContext, 2);
      });
    });

    group('.stop()', () {
      test('holding the last state', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 5,
          states: {
            FooState.Foo: b.state(immediate: [b.immediate(FooState.Bar)]),
            FooState.Bar: b.state(),
          },
        ))
          ..start()
          ..stop();

        expect(
            machine.status,
            isA<MachineStatusStopped<FooState, FooEvent, int>>()
                .having((s) => s.lastState.value, 'last value ', FooState.Bar));
        expect(
            machine.status,
            isA<MachineStatusStopped<FooState, FooEvent, int>>()
                .having((s) => s.lastState.context, 'last  context', 5));
      });

      test('will execute exit effects', () {
        final b = MachineBuilder<FooState, FooEvent, int>();
        final mock = MockEffectExitMock();
        final mock2 = MockEffectExitMock();

        b.machine(b.config(
          initial: FooState.Foo,
          context: 5,
          states: {
            FooState.Foo: b.state(effect: [
              b.effect.call((_) => mock),
              b.effect.call((_) => mock2),
            ]),
          },
        ))
          ..start()
          ..stop();

        verify(mock.call(any)).called(1);
        verify(mock2.call(any)).called(1);
      });
    });

    group('effects', () {
      test('execute callbacks on entering and on leaving state', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final fooEnter = MockEffectEnterMock();
        final fooExit = MockEffectExitMock();
        final barEnter = MockEffectEnterMock();
        when(fooEnter.call(any)).thenReturn(fooExit);
        when(barEnter.call(any)).thenReturn(null);

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              transition: [b.onEvent<_Next>(FooState.Bar)],
              effect: [b.effect.call(fooEnter)],
            ),
            FooState.Bar: b.state(
              effect: [b.effect.call(barEnter)],
            ),
          },
        ))
          ..start();

        machine.send(const FooEvent.Next());

        verifyInOrder([fooEnter(any), fooExit(any), barEnter(any)]);
      });

      test('can send events to machine', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              transition: [b.onEvent<_Next>(FooState.Bar)],
              effect: [
                b.effect.call((effect) {
                  effect.send(const FooEvent.Next());
                })
              ],
            ),
            FooState.Bar: b.state(),
          },
        ))
          ..start();

        expect(machine.debugStateValue, FooState.Bar);
      });

      test('can no longer send events to machine after state has been left',
          () async {
        final b = MachineBuilder<FooState, FooEvent, int>();
        final c = Completer<void>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              transition: [
                b.onEvent<_ToBar>(FooState.Bar),
                b.onEvent<_ToBaz>(FooState.Baz),
              ],
              effect: [
                b.effect.call((effect) {
                  c.future.then((value) {
                    effect.send(const FooEvent.ToBaz());
                  });

                  return (_) => c.complete();
                })
              ],
            ),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start()
          ..send(const FooEvent.ToBar());

        await c.future;

        expect(machine.debugStateValue, FooState.Bar);
      });

      test('enter callback has correct effect value', () {
        final b = MachineBuilder<FooState, FooEvent, int>();
        late EffectValue<FooState, FooEvent, int> e;

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              transition: [
                b.onEvent<_ToBar>(FooState.Bar),
                b.onEvent<_ToBaz>(FooState.Baz),
              ],
              effect: [
                b.effect.call((effect) {
                  e = effect;
                })
              ],
            ),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        expect(
            e,
            isA<EffectValue<FooState, FooEvent, int>>()
                .having((s) => s.isActive, 'isActive while in state', isTrue));

        expect(
            e,
            isA<EffectValue<FooState, FooEvent, int>>().having(
                (s) => s.getState().value, 'state value', FooState.Foo));

        expect(
            e,
            isA<EffectValue<FooState, FooEvent, int>>()
                .having((s) => s.getState().context, 'context', 0));

        machine.send(const FooEvent.ToBar());
        expect(
            e,
            isA<EffectValue<FooState, FooEvent, int>>().having(
                (s) => s.isActive,
                'isActive while no longer in state',
                isFalse));
      });

      test('exit callback has correct value', () {
        final b = MachineBuilder<FooState, FooEvent, int>();
        late MachineState<FooState, FooEvent, int> state;

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              transition: [
                b.onEvent<_ToBar>(FooState.Bar, reduce: (ctx, _) => ctx + 1),
              ],
              reduceOnExit: (ctx) => ctx + 1,
              effect: [
                b.effect.call((_) {
                  return (s) => state = s;
                })
              ],
            ),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        machine.send(const FooEvent.ToBar());
        expect(
            state,
            isA<MachineState<FooState, FooEvent, int>>()
                .having((s) => s.value, 'value', FooState.Foo));

        expect(
            state,
            isA<MachineState<FooState, FooEvent, int>>()
                .having((s) => s.context, 'context', 2));
      });
    });

    group('transition', () {
      test('will transition to correct state', () {
        final b = MachineBuilder<FooState, FooEvent, void>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: null,
          states: {
            FooState.Foo: b.state(transition: [b.onEvent<_Next>(FooState.Bar)]),
            FooState.Bar: b.state(transition: [b.onEvent<_Next>(FooState.Baz)]),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        machine.send(FooEvent.Next());
        expect(machine.debugStateValue, FooState.Bar);

        machine.send(FooEvent.Next());
        expect(machine.debugStateValue, FooState.Baz);
      });

      test('will result in correct context', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(transition: [
              b.onEvent<_Next>(FooState.Bar, reduce: (ctx, _) => ctx + 1)
            ]),
            FooState.Bar: b.state(transition: [
              b.onEvent<_Next>(FooState.Baz, reduce: (ctx, _) => ctx + 1)
            ]),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        machine.send(const FooEvent.Next());
        expect(machine.debugStateContext, 1);

        machine.send(const FooEvent.Next());
        expect(machine.debugStateContext, 2);
      });

      test('will use event with data', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              transition: [
                b.onEvent<_Next>(FooState.Baz),
                b.onEvent<_NextWithData>(FooState.Baz,
                    reduce: (ctx, event) => ctx + event.data),
              ],
            ),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        machine.send(FooEvent.NextWithData(100));
        expect(machine.debugStateContext, 100);
      });

      test('will prevent transition with condition', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(transition: [
              b.onEvent<_Next>(FooState.Bar, condition: (_, __) => false),
              b.onEvent<_Next>(FooState.Baz),
            ]),
            FooState.Bar: b.state(),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        machine.send(const FooEvent.Next());
        expect(machine.debugStateValue, FooState.Baz);
      });
    });

    group('immediate', () {
      test('transition on initial state', () {
        final b = MachineBuilder<FooState, FooEvent, void>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: null,
          states: {
            FooState.Foo: b.state(immediate: [b.immediate(FooState.Bar)]),
            FooState.Bar: b.state(immediate: [b.immediate(FooState.Baz)]),
            FooState.Baz: b.state(),
          },
        ))
          ..start();
        expect(machine.debugStateValue, FooState.Baz);
      });

      test('can transition after event', () {
        final b = MachineBuilder<FooState, FooEvent, void>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: null,
          states: {
            FooState.Foo: b.state(transition: [b.onEvent<_Next>(FooState.Bar)]),
            FooState.Bar: b.state(immediate: [b.immediate(FooState.Baz)]),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        machine.send(const FooEvent.Next());
        expect(machine.debugStateValue, FooState.Baz);
      });

      test('will result in correct context', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              immediate: [b.immediate(FooState.Bar, reduce: (ctx) => ctx + 1)],
            ),
            FooState.Bar: b.state(
              immediate: [b.immediate(FooState.Baz, reduce: (ctx) => ctx + 1)],
            ),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        expect(machine.debugStateContext, 2);
      });

      test('prevent transition with condition', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              immediate: [b.immediate(FooState.Bar)],
            ),
            FooState.Bar: b.state(
              immediate: [b.immediate(FooState.Baz, condition: (ctx) => false)],
            ),
            FooState.Baz: b.state(),
          },
        ))
          ..start();

        expect(machine.debugStateValue, FooState.Bar);
      });
    });

    group('context', () {
      test('has correct initial value', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 100,
          states: {
            FooState.Foo: b.state(),
          },
        ))
          ..start();
        expect(machine.debugStateContext, 100);
      });

      test('has correct initial value on immediate transition', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(immediate: [
              b.immediate(FooState.Bar, reduce: (ctx) => ctx + 100)
            ]),
            FooState.Bar: b.state(),
          },
        ))
          ..start();
        expect(machine.debugStateContext, 100);
      });

      test('can be updated by reduceOnEnter/Exit', () {
        final b = MachineBuilder<FooState, FooEvent, int>();

        final machine = b.machine(b.config(
          initial: FooState.Foo,
          context: 0,
          states: {
            FooState.Foo: b.state(
              transition: [b.onEvent<_Next>(FooState.Bar)],
              reduceOnEnter: (ctx) => ctx + 100,
              reduceOnExit: (ctx) => ctx + 100,
            ),
            FooState.Bar: b.state(
              reduceOnEnter: (ctx) => ctx + 100,
            ),
          },
        ))
          ..start();

        machine.send(const FooEvent.Next());
        expect(machine.debugStateContext, 300);
      });
    });
  });
}
