import 'package:flutter_endlich/flutter_endlich.dart';
import 'package:flutter_endlich/src/machine.dart';
import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';

import 'managed_test.mocks.dart';
part 'managed_test.freezed.dart';

@freezed
class FooEvent with _$FooEvent {
  const factory FooEvent.Next() = _Next;
  const factory FooEvent.NextWithData(int data) = _NextWithData;
  const factory FooEvent.ToFoo() = _ToFoo;
  const factory FooEvent.ToBar() = _ToBar;
  const factory FooEvent.ToBaz() = _ToBaz;
}

enum FooState { Foo, Bar, Baz }

class EffectEnterMock<T> extends Mock {
  EffectLeave<FooState, FooEvent, T>? call(
      EffectValue<FooState, FooEvent, T> effect);
}

class EffectExitMock<T> extends Mock {
  void call(MachineState<FooState, FooEvent, T> state);
}

class MachineListener extends Mock {
  void call();
}

class ManagedChangeTwo<AValue, AContext, BValue, BContext> extends Mock {
  void call(
      MachineState<AValue, Object, AContext> stateA,
      MachineState<BValue, Object, BContext> stateB,
      void Function(FooEvent event) send);
}

class VerifyOne<AValue, AContext> extends Mock {
  void call(MachineState<AValue, Object, AContext> state,
      void Function(FooEvent event) send);
}

class VerifyTwo<AValue, AContext, BValue, BContext> extends Mock {
  void call(
      MachineState<AValue, Object, AContext> stateA,
      MachineState<BValue, Object, BContext> stateB,
      void Function(FooEvent event) send);
}

class VerifyThree<AValue, AContext, BValue, BContext, CValue, CContext>
    extends Mock {
  void call(
      MachineState<AValue, Object, AContext> state,
      MachineState<BValue, Object, BContext> stateB,
      MachineState<CValue, Object, CContext> stateC,
      void Function(FooEvent event) send);
}

class VerifyFour<AValue, AContext, BValue, BContext, CValue, CContext, DValue,
    DContext> extends Mock {
  void call(
      MachineState<AValue, Object, AContext> state,
      MachineState<BValue, Object, BContext> stateB,
      MachineState<CValue, Object, CContext> stateC,
      MachineState<DValue, Object, DContext> stateD,
      void Function(FooEvent event) send);
}

class VerifyFive<AValue, AContext, BValue, BContext, CValue, CContext, DValue,
    DContext, EValue, EContext> extends Mock {
  void call(
      MachineState<AValue, Object, AContext> state,
      MachineState<BValue, Object, BContext> stateB,
      MachineState<CValue, Object, CContext> stateC,
      MachineState<DValue, Object, DContext> stateD,
      MachineState<EValue, Object, EContext> stateE,
      void Function(FooEvent event) send);
}

class ContextA {}

class ContextB {}

class ContextC {}

class ContextD {}

class ContextE {}

@GenerateMocks([
  MachineListener,
  EffectEnterMock,
  EffectExitMock,
  ManagedChangeTwo,
  VerifyOne,
  VerifyTwo,
  VerifyThree,
  VerifyFour,
  VerifyFive
])
void main() {
  test('managed machines can be nested into other machine', () {
    final deepB = MachineBuilder<FooState, FooEvent, int>();
    final deeperB = MachineBuilder<FooState, FooEvent, String>();
    final deepestB = MachineBuilder<FooState, FooEvent, DateTime>();

    final deepestMachine = deepestB.managed(deepestB.config(
      initial: FooState.Foo,
      context: DateTime.now(),
      states: {
        FooState.Foo: deepestB.state(),
        FooState.Bar: deepestB.state(),
      },
    ));
    final deepest2Machine = deepestB.managed(deepestB.config(
      initial: FooState.Foo,
      context: DateTime.now(),
      states: {
        FooState.Foo:
            deepestB.state(transition: [deepestB.onEvent<_Next>(FooState.Bar)]),
        FooState.Bar: deepestB.state(),
      },
    ));

    final deeperMachine = deeperB.managed(deeperB.config(
      initial: FooState.Foo,
      context: 'hello',
      states: {
        FooState.Foo: deeperB.state.atomic(
          transition: [deeperB.onEvent<_Next>(FooState.Bar)],
          managed: deeperB.state.manage.one(machine: deepestMachine),
        ),
        FooState.Bar: deeperB.state(
          transition: [deeperB.onEvent<_Next>(FooState.Foo)],
        ),
      },
    ));

    final deeper2Machine = deeperB.managed(deeperB.config(
      initial: FooState.Foo,
      context: 'world',
      states: {
        FooState.Foo: deeperB.state.atomic(
          transition: [deeperB.onEvent<_Next>(FooState.Bar)],
          managed: deeperB.state.manage.one(machine: deepest2Machine),
        ),
        FooState.Bar: deeperB.state(
          transition: [deeperB.onEvent<_Next>(FooState.Foo)],
        ),
      },
    ));

    final deepMachine = deepB.machine(deepB.config(
      initial: FooState.Foo,
      context: 0,
      states: {
        FooState.Foo: deepB.state.atomic(
          transition: [deepB.onEvent<_Next>(FooState.Bar)],
          managed: deepB.state.manage.two(
            machineA: deeperMachine,
            machineB: deeper2Machine,
          ),
        ),
        FooState.Bar: deepB.state(
          transition: [deepB.onEvent<_Next>(FooState.Foo)],
        ),
      },
    ));

    expect(deepMachine.status, isA<MachineStatusNotStarted>());
    expect(deeperMachine.status, isA<MachineStatusNotStarted>());
    expect(deeper2Machine.status, isA<MachineStatusNotStarted>());
    expect(deepestMachine.status, isA<MachineStatusNotStarted>());
    expect(deepest2Machine.status, isA<MachineStatusNotStarted>());

    deepMachine.start();
    expect(
        deepMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeperMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeper2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepestMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepest2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));

    deepMachine.stop();
    expect(
        deepMachine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deeperMachine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deeper2Machine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deepestMachine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deepest2Machine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));

    deepMachine.start();
    expect(
        deepMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeperMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeper2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepestMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepest2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));

    deepMachine.send(const FooEvent.Next());
    expect(
        deepMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Bar));
    expect(
        deeperMachine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deeper2Machine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deepestMachine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deepest2Machine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));

    deepMachine.send(const FooEvent.Next());
    expect(
        deepMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeperMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeper2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepestMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepest2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));

    deeperMachine.send(const FooEvent.Next());
    expect(
        deepMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeperMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Bar));
    expect(
        deeper2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepestMachine.status,
        isA<MachineStatusStopped>()
            .having((s) => s.lastState.value, 'value', FooState.Foo));
    expect(
        deepest2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));

    deeperMachine.send(const FooEvent.Next());
    expect(
        deepMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeperMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deeper2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepestMachine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));
    expect(
        deepest2Machine.status,
        isA<MachineStatusRunning>()
            .having((s) => s.state.value, 'value', FooState.Foo));

    deepMachine.dispose();
    expect(deepMachine.status, isA<MachineStatusStopped>());
    expect(deeperMachine.status, isA<MachineStatusStopped>());
    expect(deeper2Machine.status, isA<MachineStatusStopped>());
    expect(deepestMachine.status, isA<MachineStatusStopped>());
    expect(deepest2Machine.status, isA<MachineStatusStopped>());
  });

  test('managed will immediate transition', () {
    final deepB = MachineBuilder<FooState, FooEvent, int>();
    final deeperB = MachineBuilder<FooState, FooEvent, String>();
    final deepestB = MachineBuilder<FooState, FooEvent, DateTime>();

    final deepestMachine = deepestB.managed(deepestB.config(
      initial: FooState.Foo,
      context: DateTime.now(),
      states: {
        FooState.Foo:
            deepestB.state(immediate: [deepestB.immediate(FooState.Bar)]),
        FooState.Bar: deepestB.state(),
      },
    ));

    final deeperMachine = deeperB.managed(deeperB.config(
      initial: FooState.Foo,
      context: 'hello',
      states: {
        FooState.Foo:
            deeperB.state(immediate: [deeperB.immediate(FooState.Bar)]),
        FooState.Bar: deeperB.state.atomic(
          managed: deeperB.state.manage.one(machine: deepestMachine),
        )
      },
    ));

    final deepMachine = deepB.machine(deepB.config(
      initial: FooState.Foo,
      context: 0,
      states: {
        FooState.Foo: deepB.state(immediate: [deepB.immediate(FooState.Bar)]),
        FooState.Bar: deepB.state.atomic(
          managed: deepB.state.manage.one(machine: deeperMachine),
        ),
      },
    ))
      ..start();

    expect(deepMachine.debugStateValue, FooState.Bar);
    expect(deeperMachine.debugStateValue, FooState.Bar);
    expect(deepestMachine.debugStateValue, FooState.Bar);
  });

  test(
      'managed machines will have their state updated before the first listener is called',
      () {
    final deepB = MachineBuilder<FooState, FooEvent, int>();
    final deeperB = MachineBuilder<FooState, FooEvent, String>();
    final deepestB = MachineBuilder<FooState, FooEvent, DateTime>();

    final deepestMachine = deepestB.managed(deepestB.config(
      initial: FooState.Foo,
      context: DateTime.now(),
      states: {
        FooState.Foo:
            deepestB.state(immediate: [deepestB.immediate(FooState.Bar)]),
        FooState.Bar: deepestB.state(),
      },
    ));

    final deeperMachine = deeperB.managed(deeperB.config(
      initial: FooState.Foo,
      context: 'hello',
      states: {
        FooState.Foo:
            deeperB.state(immediate: [deeperB.immediate(FooState.Bar)]),
        FooState.Bar: deeperB.state
            .atomic(managed: deeperB.state.manage.one(machine: deepestMachine)),
      },
    ));

    final deepMachine = deepB.machine(deepB.config(
      initial: FooState.Foo,
      context: 0,
      states: {
        FooState.Foo: deepB.state(immediate: [deepB.immediate(FooState.Bar)]),
        FooState.Bar: deepB.state.atomic(
          managed: deepB.state.manage.one(machine: deeperMachine),
        ),
      },
    ));

    deepMachine.addListener(() {
      expect(deepMachine.debugStateValue, FooState.Bar);
      expect(deeperMachine.debugStateValue, FooState.Bar);
      expect(deepestMachine.debugStateValue, FooState.Bar);
    });

    deepMachine.start();
  });

  test('listeners are called in correct order', () {
    final builder = MachineBuilder<FooState, FooEvent, int>();
    ManagedMachine<FooState, FooEvent, int> managed(
        [ManagedMachine<FooState, FooEvent, int>? manageA,
        ManagedMachine<FooState, FooEvent, int>? manageB]) {
      return builder.managed(builder.config(
        initial: FooState.Foo,
        context: 0,
        states: {
          FooState.Foo: null != manageA
              ? builder.state.atomic(
                  managed: builder.state.manage
                      .two(machineA: manageA, machineB: manageB!))
              : builder.state(),
          FooState.Bar: builder.state(),
        },
      ));
    }

    final m11 = managed();
    final m10 = managed();
    final m9 = managed();
    final m8 = managed();
    final m7 = managed(m10, m11);
    final m6 = managed();
    final m5 = managed();
    final m4 = managed(m8, m9);
    final m3 = managed(m6, m7);
    final m2 = managed(m4, m5);
    final m1 = builder.machine(builder.config(
      initial: FooState.Foo,
      context: 0,
      states: {
        FooState.Foo: builder.state.atomic(
            managed: builder.state.manage.two(machineA: m2, machineB: m3)),
        FooState.Bar: builder.state(),
      },
    ));

    final m11L = MachineListener();
    final m10L = MachineListener();
    final m9L = MachineListener();
    final m8L = MachineListener();
    final m7L = MachineListener();
    final m6L = MachineListener();
    final m5L = MachineListener();
    final m4L = MachineListener();
    final m3L = MachineListener();
    final m2L = MachineListener();
    final m1L = MachineListener();

    m3.addListener(m3L);
    m4.addListener(m4L);
    m5.addListener(m5L);
    m11.addListener(m11L);
    m10.addListener(m10L);
    m6.addListener(m6L);
    m7.addListener(m7L);
    m8.addListener(m8L);
    m9.addListener(m9L);
    m1.addListener(m1L);
    m2.addListener(m2L);

    m1.start();
    verifyInOrder([
      m1L(),
      m2L(),
      m3L(),
      m4L(),
      m5L(),
      m6L(),
      m7L(),
      m8L(),
      m9L(),
      m10L(),
      m11L(),
    ]);
  });

  test('execute effects in correct order', () {
    final deepB = MachineBuilder<FooState, FooEvent, int>();
    final deeperB = MachineBuilder<FooState, FooEvent, String>();
    final deepestB = MachineBuilder<FooState, FooEvent, DateTime>();

    final deepEnterMock = MockEffectEnterMock<int>();
    final deeperEnterMock = MockEffectEnterMock<String>();
    final deepestEnterMock = MockEffectEnterMock<DateTime>();

    final deepExitMock = MockEffectExitMock<int>();
    final deeperExitMock = MockEffectExitMock<String>();
    final deepestExitMock = MockEffectExitMock<DateTime>();

    when(deepEnterMock.call(any)).thenReturn(deepExitMock);
    when(deeperEnterMock.call(any)).thenReturn(deeperExitMock);
    when(deepestEnterMock.call(any)).thenReturn(deepestExitMock);

    final deepestMachine = deepestB.managed(deepestB.config(
      initial: FooState.Foo,
      context: DateTime.now(),
      states: {
        FooState.Foo: deepestB.state(
          effect: [deepestB.effect.create(deepestEnterMock)],
        ),
        FooState.Bar: deepestB.state(),
      },
    ));

    final deeperMachine = deeperB.managed(deeperB.config(
      initial: FooState.Foo,
      context: 'hello',
      states: {
        FooState.Foo: deeperB.state.atomic(
            transition: [deeperB.onEvent<_Next>(FooState.Bar)],
            effect: [deeperB.effect.create(deeperEnterMock)],
            managed: deeperB.state.manage.one(machine: deepestMachine)),
        FooState.Bar: deeperB.state(
          transition: [deeperB.onEvent<_Next>(FooState.Foo)],
        ),
      },
    ));

    final deepMachine = deepB.machine(deepB.config(
      initial: FooState.Foo,
      context: 0,
      states: {
        FooState.Foo: deepB.state.atomic(
            transition: [deepB.onEvent<_Next>(FooState.Bar)],
            effect: [deepB.effect.create(deepEnterMock)],
            managed: deepB.state.manage.one(
              machine: deeperMachine,
            )),
        FooState.Bar: deepB.state(
          transition: [deepB.onEvent<_Next>(FooState.Foo)],
        ),
      },
    ))
      ..start();

    deepMachine..send(const FooEvent.Next())..send(const FooEvent.Next());
    verifyInOrder([
      deepEnterMock(any),
      deeperEnterMock(any),
      deepestEnterMock(any),

      // exit
      deepestExitMock(any),
      deeperExitMock(any),
      deepExitMock(any),

      // reentry
      deepEnterMock(any),
      deeperEnterMock(any),
      deepestEnterMock(any),
    ]);

    reset(deepEnterMock);
    reset(deeperEnterMock);
    reset(deepestEnterMock);
    reset(deepExitMock);
    reset(deeperExitMock);
    reset(deepestExitMock);

    when(deepEnterMock.call(any)).thenReturn(deepExitMock);
    when(deeperEnterMock.call(any)).thenReturn(deeperExitMock);
    when(deepestEnterMock.call(any)).thenReturn(deepestExitMock);

    verifyNoMoreInteractions(deepEnterMock);
    verifyNoMoreInteractions(deepExitMock);

    deeperMachine..send(const FooEvent.Next())..send(const FooEvent.Next());
    verifyInOrder([
      deepestExitMock(any),
      deeperExitMock(any),

      // reentry
      deeperEnterMock(any),
      deepestEnterMock(any),
    ]);
  });

  test('start managed in different value or context', () {
    final rootB = MachineBuilder<FooState, FooEvent, int>();
    final level1B = MachineBuilder<FooState, FooEvent, String>();

    final level1Machine = level1B.managed(level1B.config(
      initial: FooState.Foo,
      context: 'hello',
      states: {
        FooState.Foo: level1B.state(),
        FooState.Bar: level1B.state(),
      },
    ));

    rootB.machine(rootB.config(
      initial: FooState.Foo,
      context: 0,
      states: {
        FooState.Foo: rootB.state.atomic(
          managed: rootB.state.manage.one<FooState, String>(
            machine: level1Machine,
            value: (_) => FooState.Bar,
            context: (_) => 'world',
          ),
        ),
      },
    ))
      ..start();

    expect(level1Machine.debugStateValue, FooState.Bar);
    expect(level1Machine.debugStateContext, 'world');
  });

  test('listen to changes in managed machine', () {
    final rootB = MachineBuilder<FooState, FooEvent, int>();
    final level1A = MachineBuilder<FooState, FooEvent, String>();
    final level1B = MachineBuilder<FooState, FooEvent, DateTime>();

    final now = DateTime.now();
    final january = DateTime(DateTime.january);

    final changeTwo =
        MockManagedChangeTwo<FooState, String, FooState, DateTime>();

    final level1AMachine = level1A.managed(level1A.config(
      initial: FooState.Foo,
      context: 'hello',
      states: {
        FooState.Foo: level1A.state(transition: [
          level1A.onEvent<_Next>(FooState.Bar, reduce: (_, __) => 'hello'),
        ]),
        FooState.Bar: level1A.state(),
      },
    ));

    final level1BMachine = level1B.managed(level1B.config(
      initial: FooState.Foo,
      context: now,
      states: {
        FooState.Foo: level1B.state(),
        FooState.Bar: level1B.state(
          transition: [
            level1B.onEvent<_Next>(FooState.Foo, reduce: (_, __) => january)
          ],
        ),
      },
    ));

    final rootMachine = rootB.machine(rootB.config(
      initial: FooState.Foo,
      context: 0,
      states: {
        FooState.Foo: rootB.state.atomic(
          managed: rootB.state.manage.two<FooState, String, FooState, DateTime>(
            machineA: level1AMachine,
            machineB: level1BMachine,
            contextA: (_) => 'world',
            valueB: (_) => FooState.Bar,
            anyChange: changeTwo,
          ),
        ),
      },
    ))
      ..start();

    level1AMachine.send(const FooEvent.Next());
    level1BMachine.send(const FooEvent.Next());

    verifyInOrder([
      changeTwo(
        argThat(equals(MachineState<FooState, FooEvent, String>(
            value: FooState.Foo, context: 'world'))),
        argThat(equals(MachineState<FooState, FooEvent, DateTime>(
            value: FooState.Bar, context: now))),
        argThat(equals(rootMachine.send)),
      ),
      changeTwo(
        argThat(equals(MachineState<FooState, FooEvent, String>(
            value: FooState.Bar, context: 'hello'))),
        argThat(equals(MachineState<FooState, FooEvent, DateTime>(
            value: FooState.Bar, context: now))),
        argThat(equals(rootMachine.send)),
      ),
      changeTwo(
        argThat(equals(MachineState<FooState, FooEvent, String>(
            value: FooState.Bar, context: 'hello'))),
        argThat(equals(MachineState<FooState, FooEvent, DateTime>(
            value: FooState.Foo, context: january))),
        argThat(equals(rootMachine.send)),
      )
    ]);
  });

  group('builder', () {
    final contextA = ContextA();
    final contextB = ContextB();
    final contextC = ContextC();
    final contextD = ContextD();
    final contextE = ContextE();
    late MachineBuilder<FooState, FooEvent, void> rootB;
    late ManagedMachine<FooState, FooEvent, ContextA> aMachine;
    late ManagedMachine<FooState, FooEvent, ContextB> bMachine;
    late ManagedMachine<FooState, FooEvent, ContextC> cMachine;
    late ManagedMachine<FooState, FooEvent, ContextD> dMachine;
    late ManagedMachine<FooState, FooEvent, ContextE> eMachine;

    setUp(() {
      rootB = MachineBuilder<FooState, FooEvent, void>();
      final a = MachineBuilder<FooState, FooEvent, ContextA>();
      final b = MachineBuilder<FooState, FooEvent, ContextB>();
      final c = MachineBuilder<FooState, FooEvent, ContextC>();
      final d = MachineBuilder<FooState, FooEvent, ContextD>();
      final e = MachineBuilder<FooState, FooEvent, ContextE>();
      aMachine = a.managed(a.config(
        initial: FooState.Foo,
        context: contextA,
        states: {
          FooState.Foo: a.state(),
        },
      ));
      bMachine = b.managed(b.config(
        initial: FooState.Foo,
        context: contextB,
        states: {
          FooState.Foo: b.state(),
        },
      ));
      cMachine = c.managed(c.config(
        initial: FooState.Foo,
        context: contextC,
        states: {
          FooState.Foo: c.state(),
        },
      ));
      dMachine = d.managed(d.config(
        initial: FooState.Foo,
        context: contextD,
        states: {
          FooState.Foo: d.state(),
        },
      ));
      eMachine = e.managed(e.config(
        initial: FooState.Foo,
        context: contextE,
        states: {
          FooState.Foo: e.state(),
        },
      ));
    });
    test('.one()', () {
      final one = MockVerifyOne<FooState, ContextA>();

      final rootMachine = rootB.machine(rootB.config(
        initial: FooState.Foo,
        context: 0,
        states: {
          FooState.Foo: rootB.state.atomic(
            managed: rootB.state.manage.one(
              machine: aMachine,
              change: one,
            ),
          ),
        },
      ))
        ..start();

      verify(one(
        argThat(equals(MachineState<FooState, FooEvent, ContextA>(
            value: FooState.Foo, context: contextA))),
        argThat(equals(rootMachine.send)),
      )).called(1);
    });

    test('.two()', () {
      final two = MockVerifyTwo<FooState, ContextA, FooState, ContextB>();

      final rootMachine = rootB.machine(rootB.config(
        initial: FooState.Foo,
        context: 0,
        states: {
          FooState.Foo: rootB.state.atomic(
            managed: rootB.state.manage.two(
              machineA: aMachine,
              machineB: bMachine,
              anyChange: two,
            ),
          ),
        },
      ))
        ..start();

      verify(two(
        argThat(equals(MachineState<FooState, FooEvent, ContextA>(
            value: FooState.Foo, context: contextA))),
        argThat(equals(MachineState<FooState, FooEvent, ContextB>(
            value: FooState.Foo, context: contextB))),
        argThat(equals(rootMachine.send)),
      )).called(1);
    });

    test('.three()', () {
      final three = MockVerifyThree<FooState, ContextA, FooState, ContextB,
          FooState, ContextC>();

      final rootMachine = rootB.machine(rootB.config(
        initial: FooState.Foo,
        context: 0,
        states: {
          FooState.Foo: rootB.state.atomic(
            managed: rootB.state.manage.three(
              machineA: aMachine,
              machineB: bMachine,
              machineC: cMachine,
              anyChange: three,
            ),
          ),
        },
      ))
        ..start();

      verify(three(
        argThat(equals(MachineState<FooState, FooEvent, ContextA>(
            value: FooState.Foo, context: contextA))),
        argThat(equals(MachineState<FooState, FooEvent, ContextB>(
            value: FooState.Foo, context: contextB))),
        argThat(equals(MachineState<FooState, FooEvent, ContextC>(
            value: FooState.Foo, context: contextC))),
        argThat(equals(rootMachine.send)),
      )).called(1);
    });

    test('.four()', () {
      final four = MockVerifyFour<FooState, ContextA, FooState, ContextB,
          FooState, ContextC, FooState, ContextD>();

      final rootMachine = rootB.machine(rootB.config(
        initial: FooState.Foo,
        context: 0,
        states: {
          FooState.Foo: rootB.state.atomic(
            managed: rootB.state.manage.four(
              machineA: aMachine,
              machineB: bMachine,
              machineC: cMachine,
              machineD: dMachine,
              anyChange: four,
            ),
          ),
        },
      ))
        ..start();

      verify(four(
        argThat(equals(MachineState<FooState, FooEvent, ContextA>(
            value: FooState.Foo, context: contextA))),
        argThat(equals(MachineState<FooState, FooEvent, ContextB>(
            value: FooState.Foo, context: contextB))),
        argThat(equals(MachineState<FooState, FooEvent, ContextC>(
            value: FooState.Foo, context: contextC))),
        argThat(equals(MachineState<FooState, FooEvent, ContextD>(
            value: FooState.Foo, context: contextD))),
        argThat(equals(rootMachine.send)),
      )).called(1);
    });

    test('.five()', () {
      final five = MockVerifyFive<FooState, ContextA, FooState, ContextB,
          FooState, ContextC, FooState, ContextD, FooState, ContextE>();

      final rootMachine = rootB.machine(rootB.config(
        initial: FooState.Foo,
        context: 0,
        states: {
          FooState.Foo: rootB.state.atomic(
            managed: rootB.state.manage.five(
              machineA: aMachine,
              machineB: bMachine,
              machineC: cMachine,
              machineD: dMachine,
              machineE: eMachine,
              anyChange: five,
            ),
          ),
        },
      ))
        ..start();

      verify(five(
        argThat(equals(MachineState<FooState, FooEvent, ContextA>(
            value: FooState.Foo, context: contextA))),
        argThat(equals(MachineState<FooState, FooEvent, ContextB>(
            value: FooState.Foo, context: contextB))),
        argThat(equals(MachineState<FooState, FooEvent, ContextC>(
            value: FooState.Foo, context: contextC))),
        argThat(equals(MachineState<FooState, FooEvent, ContextD>(
            value: FooState.Foo, context: contextD))),
        argThat(equals(MachineState<FooState, FooEvent, ContextE>(
            value: FooState.Foo, context: contextE))),
        argThat(equals(rootMachine.send)),
      )).called(1);
    });
  });
}
